"""
Author: Arief Anbiya
Email: anbarief@live.com
Date: 2 Sept 2020

The dataset used is the Federal Interest Rates, collected from:

https://www.kaggle.com/federalreserve/interest-rates

The function plot_rates can plot the effective fund rates, unemployment rates, and inflation rates.
To plot their correlations, use the function correlation_plot.

DATASET FILE: "federal_interest_rates_kaggle.csv"
Download the dataset from Google Drive link
(see https://anbarief.github.io/actuarydesk/)
"""

import math

import pandas
import matplotlib.pyplot as plt

FILE_PATH = "PATH OF THE DATASET"

df = pandas.read_csv(FILE_PATH)
n = len(df)
time_points = []
for i in range(n):
    year = df['Year'][i]
    month = df['Month'][i]/12
    day = (df['Day'][i]/(31))/12
    time_point = year + month + day
    time_points.append(time_point)
df['Time Points'] = time_points

labels = ['Effective Federal Funds Rate', \
              'Unemployment Rate', \
              'Inflation Rate']
rates = [df[i] for i in labels]
index = [0,1,2]
color = ['red', 'green', 'blue']

def plot_rates(min_year, max_year, effective_rate = True, \
               unemployment_rate = False, inflation_rate = False):

    bools = [effective_rate, unemployment_rate, inflation_rate]
    if not any(bools):
        return None
    n_true = len([b for b in bools if b])
    
    if n_true==3:
        pass
    elif n_true==2:
        if not effective_rate:
            index.pop(0)
        if not unemployment_rate:
            index.pop(1)
        if not inflation_rate:
            index.pop(2)
    else:
        if effective_rate:
            index = [0]
        if unemployment_rate:
            index = [1]
        if inflation_rate:
            index = [2]
        
    rates_index = []
    for i in range(n):    
        rate = [rates[idx][i] for idx in index]
        if all([not math.isnan(r) for r in rate]):
            rates_index.append(i)
            
    fig, ax = plt.subplots()
    filtered_index = [i for i in rates_index if min_year <= time_points[i] <= max_year]
    for idx in index:
        line = ax.plot([time_points[i] for i in filtered_index], [rates[idx][i] for i in filtered_index], '-', color=color[idx])
        line[0].set_label(labels[idx])   
    ax.legend()
    ax.set_xticks(range(min_year, max_year+1))
    ax.set_xticklabels(range(min_year, max_year+1), rotation = -90, fontsize=7)
    ax.set_ylabel("Rate")
    fig.show()
    return None

def correlation_plot(min_year, max_year, var_1, var_2, var_3 = False):

    rates_index = []
    x = df[var_1]; y = df[var_2]
    fig, ax = plt.subplots()

    if not var_3:
        for i in range(n):
            rate_1 = x[i]; rate_2 = y[i]
            if (not math.isnan(rate_1)) \
               and (not math.isnan(rate_2)):
                rates_index.append(i)
        filtered_index = [i for i in rates_index if min_year <= time_points[i] <= max_year]
        ax.plot([x[i] for i in filtered_index], [y[i] for i in filtered_index], 'o')
    else:
        color_var = [i for i in labels if ((i != var_1) and (i != var_2))][0]
        z = df[color_var]
        for i in range(n):
            rate_1 = x[i]; rate_2 = y[i]; rate_3 = z[i]
            if (not math.isnan(rate_1)) \
               and (not math.isnan(rate_2)) \
               and (not math.isnan(rate_3)):
                rates_index.append(i)
        filtered_index = [i for i in rates_index if min_year <= time_points[i] <= max_year]
        z_values = [z[i] for i in filtered_index]; max_z = max(z_values)
        colors = [(1,0,0, max(i/max_z, 0.2)) for i in z_values]
        x_values = [x[i] for i in filtered_index]; y_values = [y[i] for i in filtered_index]
        for i in range(len(colors)):
            ax.plot(x_values[i], y_values[i], 'o', color = colors[i])
        ax.set_title("Solid color represents high {}.".format(color_var))  

    ax.set_xlabel(var_1); ax.set_ylabel(var_2)
    fig.show()



#version 0.2.0

samples = ['federal_interest_rates_kaggle', \
           'financial_indicators_US_stocks_kaggle', \
           'world_development_indicator_kaggle', \
           'average_interest_rates_US_treasury', \
           'stochastic_random_walk']



"""
Author: Arief Anbiya
Email: anbarief@live.com
Date: 9 Sept 2020

The dataset used in this module is the World Development Indicators by the World Bank, available on Kaggle.
Source: https://www.kaggle.com/worldbank/world-development-indicators.
Terms of Use of the dataset is written here: https://www.worldbank.org/en/about/legal/terms-of-use-for-datasets

The plot_values function results line plot(s) of an indicator form time to time of one or more countries (comparing them).
If the colors argument is set to 'default' then the color of the lines will follow the default from Matplotlib.

If plot_values shows dynamics, the plot_bars function results bar plot of an indicator of one or more countries in a specific year.
The user can interact with the plot (by hovering on the axes).

DATASET FILE:
-"wdi_indicators_kaggle.csv" (FILE_PATH_1)
-"wdi_countries_kaggle.csv" (FILE_PATH_2)
Download the dataset from Google Drive link
(see https://anbarief.github.io/actuarydesk/)
"""

import math

import pandas
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

FILE_PATH_1 = "PATH OF THE DATASET"
FILE_PATH_2 = "PATH OF THE DATASET"

df = pandas.read_csv(FILE_PATH_1)
df_countries  = pandas.read_csv(FILE_PATH_2)
years = sorted(list(set(df['Year'])))
indicator_names= sorted(list(set(df['IndicatorName'])))
color_code = {'no data': (0, 0, 0, 0.5), \
              'Europe & Central Asia': (0, 0, 1, 0.5), \
              'Middle East & North Africa': (186/255, 138/255, 6/255, 0.5), \
              'North America': (1, 0, 0, 0.5), \
              'Latin America & Caribbean': (0, 1, 0, 0.5), \
              'Sub-Saharan Africa': (1, 0, 1, 0.5), \
              'South Asia': (0, 0.5, 0.5, 0.5), \
              'East Asia & Pacific': (1, 1, 50/255, 0.5)}

def filter_by_country(country_name):
    return df[df['CountryName']==country_name]

def filter_by_indicator_country(country_name, indicator_name):
    result = filter_by_country(country_name)
    return result[result['IndicatorName']==indicator_name]

def find_countries_with_indicator(indicator):
    countries = list(set(df['CountryName']))
    result = []
    for country in countries:
        if len(filter_by_indicator_country(country, indicator))>0:
            result.append(country)
    return result

def country_exist_by_indicator(country, indicator):
    if len(filter_by_indicator_country(country, indicator))>0:
        return True
    return False

def plot_values(countries, colors, indicator):    

    dfs=[filter_by_indicator_country(country, indicator) for country in countries]
    bools = [len(dfs[i])>0 for i in range(len(countries))]
    if not any(bools):
        print("No data")
        return None
    
    all_years = []
    for df in dfs:
        all_years += list(df['Year'])
        
    min_year = min(all_years); max_year = max(all_years)

    fig, ax = plt.subplots()
    for i in range(len(countries)):
        
        if bools[i]:
            if colors == "default":
                p=ax.plot(dfs[i]['Year'], dfs[i]['Value']);
            else:
                p=ax.plot(dfs[i]['Year'], dfs[i]['Value'], color=colors[i]);
            p[0].set_label(countries[i])
    
    ax.set_title(indicator, fontsize=10)
    ax.set_xticks(range(min_year, max_year+1)); ax.set_xticklabels([str(i) for i in range(min_year, max_year+1)], rotation = -90, fontsize=6);
    ax.legend()
    plt.tight_layout()
    fig.show()

def plot_bars(indicator, year, lowest_rank = 10):

    if lowest_rank < 1:
        print("Rank must be at least 1")
        return None

    df1=df[df['IndicatorName']==indicator]
    if len(df1)==0:
        print("No data available")
        return None
    
    df2=df1[df1['Year']==year]
    if len(df2)==0:
        print("No data available")
        return None
    
    n = len(df2['CountryCode'])
    top_m = min(lowest_rank, n)

    values = [i for i in df2['Value']]
    country_code = [i for i in df2['CountryCode']]
    country_name = [i for i in df2['CountryName']]
    region = [list(df_countries[df_countries['CountryCode']==i]['Region'])[0] for i in country_code]
    for i in range(len(region)):    
        if (type(region[i]) != str) and math.isnan(region[i]):
            region[i]='no data'
        
    paired = sorted([(i, j, k, l) for i, j, k, l in zip(country_code, country_name, region, values)], key = lambda x: x[3])
    top_m_paired = paired[-top_m:]

    fig, ax = plt.subplots()
    ax.set_ylim([min(values), 1.3*max(values)])
    for i in color_code.keys():
        p=ax.plot(0,min(values)-1000,'o', color = color_code[i])
        p[0].set_label(i)
    ax.legend()
    
    for i in range(top_m):
        ax.bar(top_m-i,  paired[-(i+1)][3],color = color_code[paired[-(i+1)][2]], ec = 'black')

    dashed_line = ax.plot(0,min(values)-1000,'--', color = 'gray', lw = 1)[0]
    def on_motion(event):
        x_pos, y_pos = event.xdata, event.ydata
        if x_pos != None:
            for i in range(1, top_m+1):
                if abs(x_pos-i) < 0.5: 
                    text = top_m_paired[i-1][1] + " ({})".format(top_m_paired[i-1][3])
                    ax.set_xlabel(text, fontsize=10)
                    if y_pos > top_m_paired[i-1][3]:
                        dashed_line.set_xdata([i, i]); dashed_line.set_ydata([top_m_paired[i-1][3], y_pos])
                    else:
                        dashed_line.set_ydata([min(values)-1000, min(values)-1000])
                    fig.show()
                    break
        
    cid = fig.canvas.mpl_connect('motion_notify_event', on_motion)
    ax.set_xticks(range(1, top_m+1)); ax.set_xticklabels(["" for i in range(top_m)], fontsize=4);
    ax.set_title(indicator, fontsize=12)
    ax.set_ylabel("Value")
    plt.tight_layout()
    fig.show()



"""
Author: Arief Anbiya
Email: anbarief@live.com
Date: 3 Sept 2020

Simple Random Walk animation.
"""

import random
import statistics

import matplotlib.pyplot as plt
import matplotlib.animation as animation

def animate_randomwalk(initial_value = 0, dt = 1, h = 1, p = 0.5, n = 1000, n_update = 5, interval_between_frame = 50):

    X = [initial_value]; t = [0]    

    fig, ax = plt.subplots()
    plot1 = ax.plot(t[0], X[0], '-', color='gray'); plot1[0].set_label('X(t)')
    plot2 = ax.plot(t[0], X[0], '-', color='blue'); plot2[0].set_label('Moving mean at t')
    ax.legend()
    ax.set_xlabel("t"); ax.set_ylabel("X(t)")

    moving_mean = [X[0]]
    def animate(frame):
        plot = []
        for i in range(n_update):
            if random.uniform(0, 1) <= p:
                X.append(X[-1] + h)
            else:
                X.append(X[-1] - h)
            t.append(t[-1]+dt)
            moving_mean.append(statistics.mean(X))
            plot.append(ax.plot([t[-2], t[-1]], [X[-2], X[-1]], '-', color = 'gray', lw = 1)[0])
            plot.append(ax.plot([t[-2], t[-1]], [moving_mean[-2], moving_mean[-1]], color = 'blue', lw = 1)[0])
        return plot

    ani = animation.FuncAnimation(fig, animate, range(n), repeat=False, interval=interval_between_frame)
    plt.show()



import setuptools

setuptools.setup(\
    name='actuarydesk',\
    version='0.2.0',\
    author='Arief Anbiya',\
    author_email='anbarief@live.com',\
    description='Tools for actuary to model and analyze simple products or for actuarial students to practice.',\
    url='https://github.com/anbarief/actuarydesk',\
    packages=setuptools.find_packages(),\
    classifiers=["Programming Language :: Python :: 3",\
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",\
                 "Operating System :: OS Independent",\
                 "Topic :: Scientific/Engineering :: Mathematics",\
                 "Intended Audience :: Financial and Insurance Industry"],\
    python_requires='>=3.6', \
    install_requires=['pandas', 'matplotlib', 'sklearn'])



"""
Author: Arief Anbiya
Email: anbarief@live.com
Date: 3 Sept 2020

The dataset used is the Average Interest Rates from US Treasury Securities, collected from:

https://fiscaldata.treasury.gov/datasets/average-interest-rates-treasury-securities/

The run_animation function shows animation of the dynamics of the Average Interest Rates of each security type.

DATASET FILE: "average_interest_rates_US_treasury.csv"
Download the dataset from Google Drive link
(see https://anbarief.github.io/actuarydesk/)
"""

import math

import pandas
import matplotlib.pyplot as plt
import matplotlib.animation as animation

FILE_PATH = "PATH OF THE DATASET"

df = pandas.read_csv(FILE_PATH)
n = len(df)
description = list(df['Security Description'])
for i in range(n):
    if description[i] == "Treasury Floating Rate Note(FRN)":
       description[i] =  "Treasury Floating Rate Note (FRN)"
    if description[i] == "Treasury Inflation-Protected Securities(TIPS)":
       description[i] =  "Treasury Inflation-Protected Securities (TIPS)"
    if description[i] == "TotalMarketable":
       description[i] =  "Total Marketable"
df['Security Description'] = description       

def run_animation(interval_between_frame=100):
    
    security_description = sorted(list(set(df['Security Description'])))
    n_security = len(security_description)

    time_points = [int(i.split('-')[0]) + int(i.split('-')[1])/12 + int(i.split('-')[2])/(31*12) \
                  for i in df['Record Date']]
    df['Time Point'] = time_points
    unique_time_points = sorted(list(set(time_points)))
    n_unique_t = len(unique_time_points)

    fig, ax = plt.subplots()
    fig.set_tight_layout(True)

    def animate(frame):

        ax.cla()
        ax.set_ylim([0, 10])
        ax.set_xlim([-1, n_security+1])
        ax.set_xticks(range(n_security))
        ax.set_xticklabels(security_description, rotation = -90, fontsize = 6)
        ax.set_yticks([2, 4, 6, 8, 10])
        ax.set_yticklabels(['2%', '4%', '6%', '8%', '10%'])
        df1=df[df['Time Point']==unique_time_points[frame]]
        ax.set_title(list(df1['Record Date'])[0])
        ax.set_ylabel("Average Interest Rate")
        interest_rate = []
        bars=[]
        count = 0

        for i in security_description:
            df2 = df1[df1['Security Description']==i]['Average Interest Rate Amount']
            if len(df2)==0:
                interest_rate.append(0)
            else:
                interest_rate.append(df2[df2.index[0]])
            info = df1[df1['Security Description']==i]['Security Type Description']
            if len(info)!=0:
                info = info[info.index[0]]
            else:
                info =""
            if info == 'Marketable':
                color = "blue"
            elif info == 'Non-marketable':
                color = "red"
            else:
                color = "black"
            bar=ax.bar(count, interest_rate[-1], color = color)
            bars.append(bar[0])
            count+=1
        return bars    

    ani = animation.FuncAnimation(fig, animate, n_unique_t, repeat=False, interval=interval_between_frame)
    plt.show()



"""
Author: Arief Anbiya
Email: anbarief@live.com
Date: 9 Sept 2020

The dataset used is the US Stocks Financial Indicators, collected from:

https://www.kaggle.com/cnic92/200-financial-indicators-of-us-stocks-20142018

show_bias_variance(feature_variables, n_test, n_train) usage example: show_bias_variance(['Revenue', 'R&D Expenses'], 500, 2000) plots the Bias-Variance of the Prediction Error by Regression Tree models with max_depth from 1 to 30, using 'Revenue' and 'R&D Expenses' as the feature variables.

n_test is the number of points used for the test set (only 1 test set chosen at the beginning).
n_train is the number of points for the training sets (total of 50 sets for each degrees of freedom (max_depth value)).
When choosing n_test and n_train, look at the total number of points in the data after filtering the NaNs by filter_data.
If n is the total number of points after filter_data, then n_train <= n-n_test is required.

The variable we are predicting is "Class", which has value either 1 or 0, 1 meaning that the stock is recommended to be bought and keep for 1 year, 0 meaning the opposite.

DATASET FILE: "financial_indicators_US_stocks_kaggle.csv"
Download the dataset from Google Drive link
(see https://anbarief.github.io/actuarydesk/)
"""

import random
import math
import statistics

import pandas
import matplotlib.pyplot as plt
from sklearn.tree import DecisionTreeRegressor as DTR

FILE_PATH = "PATH OF THE DATASET"

df = pandas.read_csv(FILE_PATH)
n = len(df)

sectors = list(set(df['Sector']))
sector_idx = {sectors[i-1]: i  for i in range(1,len(sectors)+1)}
sector = [sector_idx[i] for i in df['Sector']]
df['Sector Index'] = sector

def filter_data(feature_variables):

    target_variable = "Class"
    X = []; Y = []
    variables = [list(df[i]) for i in feature_variables]
    for i in range(n):
        bools = [not math.isnan(var[i]) for var in variables] + [not math.isnan(df[target_variable][i])]
        if all(bools):
            X.append([var[i] for var in variables])
            Y.append(df[target_variable][i])
            
    return X,Y

def show_bias_variance(feature_variables, n_test, n_train):

    target_variable = 'Class'
    X, Y = filter_data(feature_variables)

    n_x = len(X)
    random_test_index = random.sample(range(n_x), n_test)

    X_test = [X[i] for i in random_test_index]
    Y_test = [[Y[i]] for i in random_test_index]

    max_df = 30
    dfs = range(1, max_df+1) #max_depth
    result  = {i: [] for i in dfs}
    train_pool_index = [i for i in range(n_x) if i not in random_test_index]

    train_indexes = []
    for i in range(50):
        train_indexes_sample = random.sample(train_pool_index, n_train)
        as_list = sorted(train_indexes_sample)
        train_indexes.append(as_list)

    for df in dfs:
        model = DTR(max_depth = df, max_features=len(feature_variables))
        prediction_errors = []
        training_errors = []
        for i in range(50):
            X_train = [X[j] for j in train_indexes[i]]
            Y_train = [[Y[j]] for j in train_indexes[i]]
            model.fit(X_train, Y_train)
            y_predict = model.predict(X_test)
            y_predict_train = model.predict(X_train) 
            mse = statistics.mean([(y_predict[j]-Y_test[j][0])**2   for j in range(n_test)])
            mse_train = statistics.mean([(y_predict_train[j]-Y_train[j][0])**2   for j in range(n_train)])
            prediction_errors.append(mse)
            training_errors.append(mse_train)

        result[df].append(prediction_errors)
        result[df].append(statistics.mean(prediction_errors))
        result[df].append(training_errors)
        result[df].append(statistics.mean(training_errors))

    fig, ax = plt.subplots()
    sort_dfs = sorted(dfs)

    for i in range(max_df):
        ax.plot(sort_dfs, [result[df][0][i] for df in sort_dfs], '-', color = (1,0.2,0,0.25), lw = 2)
        p=ax.plot(sort_dfs, [result[df][1] for df in sort_dfs], '-', color = (1,0,0,1), lw = 1.5)
    p[0].set_label("Expected Test Error Estimate")
    
    for i in range(max_df):
        ax.plot(sort_dfs, [result[df][2][i] for df in sort_dfs], '-', color = (0,0.2,1,0.25), lw = 2)
        p=ax.plot(sort_dfs, [result[df][3] for df in sort_dfs], '-', color = (0,0,1,1), lw = 1.5)
    p[0].set_label("Expected Training Error Estimate")

    ax.legend()
    ax.set_xlabel('Max Depth')
    ax.set_ylabel('Prediction Error')
    fig.show()
